clearvars -except HNMWOD* HNMWTC* k_gt; %close all; 
clc;
%MIT License, see LICENCE file

% Copyright (c) 2019 Yui Chun Leung
% Modified by Albert Argilaga, 2023, for the publication:
% Physics-Supervised GANs for the Generation of Materials obtained by
%Sacrifice Templating
%% Deep Convolutional Generative Adversarial Network
disp('FEM-GANs - A.Argilaga, 2023')
disp('_______________________________________')

%% Load Data
disp('Loading data HNMWOD0104_16bit, please wait...')
px=64; %24 32 40 48 56 64 72 80 112 184 224 %allowed image resolutions

if 0 %load database for training (or use reduced for demo)
rimg=HNMWOD0104_16bit; %3D image used as database

dsze=1.25; %downsize factor to hit real target dimensions
pxs=px*dsze; %sampling resolution 
settings.side=px;
%new (crop scan)
Ld=size(rimg,3);
Rd=size(rimg,2)/2;
dispstr=['Real dimensions L: ',num2str(Ld) ' and R: ',num2str(Rd)];
disp(dispstr)

L=Ld-2*80-pxs; %remove top and bottom 80 voxels
R=Rd-100-round(pxs*0.7071); %remove outer 100 voxels
dispstr=['Crop dimensions L: ',num2str(L) ' and R: ',num2str(R)];
disp(dispstr)
dispstr=['Pixel ratio: ',num2str(dsze)];
disp(dispstr)
fullsam=round(L*pi*R^2/pxs^3);
dispstr=['Maximum recommended database size: ',num2str(fullsam)];
disp(dispstr)
disp('Building database, please wait...')
tic

warning('off')
%projs = {'1'; '2'; '3'; '4'; '5'; '6'; '7'; '8'; '9'; '10'};
pas=round(fullsam*1.25); %%number of samples, factor over maximum recomm.
prismhd=zeros(pxs,pxs,pxs,pas);
prism=zeros(px,px,px,pas);
for i=1:pas 
    ranh=80+pxs/2+randi(L);
    ranr=randi(R);
    rant=rand*2*pi;
    prismhd(:,:,:,i) = ...
    rimg(Rd+round(ranr*cos(rant))-pxs/2+1:Rd+round(ranr*cos(rant))+pxs/2, ...
         Rd+round(ranr*sin(rant))-pxs/2+1:Rd+round(ranr*sin(rant))+pxs/2, ...
         ranh-pxs/2+1:ranh+pxs/2); %3D
    prism(:,:,:,i) = imresize3(prismhd(:,:,:,i),1/dsze);
end

toc
szprism=size(prismhd);
dispstr=['Real HD size :',num2str(szprism)];
disp(dispstr)
clearvars rimg prismhd linevec

trainX = preprocess(prism,px); %3D plane, needs changes im. sizes
size(trainX);
dispstr=['trainX size  :',num2str(ans)];
disp(dispstr)

else %load a reduced database sample for testing (48 images, 93MB)
load('smalldb.mat')
trainX=trainX10;
size(trainX);
dispstr=['reduced trainX size  :',num2str(ans)];
disp(dispstr)
end

if 0 %computation of database ground truth metrics
    msde=2; %largest side of element/voxel size in the FEM mesh
    iso= 30000; %-0.18445; %3e4;
    for j=1:size(prism,4)
        k_gt(:,:,j)=FEMmINTNV(prism(:,:,:,j),msde,0,1,i,iso); %real images
        j
    end
else
load('k_gt1776quartz.mat') 
end

clearvars prism
% figure(1) %uncomment to plot the training samples
% plot(-epsfv2D(1,:),-epsfv2D(2,:),'r+')
% xlabel('\epsilon_{xx}')
% ylabel('\epsilon_{yy}')
% title("Training samples")
% axis square

%% Settings (hyperparameters)
settings.latent_dim = 128; %any value
settings.batch_size = 8; settings.image_size = [px,px,px,1]; 
settings.lrD = 0.00005; settings.lrG = 0.00025; settings.beta1 = 0.50;
settings.beta2 = 0.999; settings.maxepochs = 200;

settings

%% Initialization
%% Generator
fcg=1/1;
fcd=1/1;
paramsGen.FCW1 = dlarray(initializeGaussian([fcg*128*round(px^3/512),...
    settings.latent_dim]));
paramsGen.FCb1 = dlarray(zeros(fcg*128*round(px^3/512),1,'single'));
paramsGen.TCW1 = dlarray(initializeGaussian([3,3,3,fcg*128,fcg*128]));
paramsGen.TCb1 = dlarray(zeros(fcg*128,1,'single'));
paramsGen.BNo1 = dlarray(zeros(fcg*128,1,'single'));
paramsGen.BNs1 = dlarray(ones(fcg*128,1,'single'));
paramsGen.TCW2 = dlarray(initializeGaussian([3,3,3,fcg*64,fcg*128]));
paramsGen.TCb2 = dlarray(zeros(fcg*64,1,'single'));
paramsGen.BNo2 = dlarray(zeros(fcg*64,1,'single'));
paramsGen.BNs2 = dlarray(ones(fcg*64,1,'single'));
paramsGen.TCW3 = dlarray(initializeGaussian([5,5,5,fcg*32,fcg*64]));
paramsGen.TCb3 = dlarray(zeros(fcg*32,1,'single'));
paramsGen.BNo3 = dlarray(zeros(fcg*32,1,'single'));
paramsGen.BNs3 = dlarray(ones(fcg*32,1,'single'));
paramsGen.CNW1 = dlarray(initializeGaussian([7,7,7,fcg*32,1]));
paramsGen.CNb1 = dlarray(zeros(1,1,'single'));
stGen.BN1 = []; stGen.BN2 = []; stGen.BN3 = [];

%% Discriminator
paramsDis.CNW1 = dlarray(initializeGaussian([7,7,7,1,fcd*16]));
paramsDis.CNb1 = dlarray(zeros(fcd*16,1,'single'));
paramsDis.CNW2 = dlarray(initializeGaussian([5,5,5,fcd*16,fcd*32]));
paramsDis.CNb2 = dlarray(zeros(fcd*32,1,'single'));
paramsDis.BNo1 = dlarray(zeros(fcd*32,1,'single'));
paramsDis.BNs1 = dlarray(ones(fcd*32,1,'single'));
paramsDis.CNW3 = dlarray(initializeGaussian([3,3,3,fcd*32,fcd*64]));
paramsDis.CNb3 = dlarray(zeros(fcd*64,1,'single'));
paramsDis.BNo2 = dlarray(zeros(fcd*64,1,'single'));
paramsDis.BNs2 = dlarray(ones(fcd*64,1,'single'));
paramsDis.CNW4 = dlarray(initializeGaussian([3,3,3,fcd*64,fcd*128]));
paramsDis.CNb4 = dlarray(zeros(fcd*128,1,'single'));
paramsDis.BNo3 = dlarray(zeros(fcd*128,1,'single'));
paramsDis.BNs3 = dlarray(ones(fcd*128,1,'single'));
paramsDis.CNW5 = dlarray(initializeGaussian([3,3,3,fcd*128,fcd*256]));
paramsDis.CNb5 = dlarray(zeros(fcd*256,1,'single'));
paramsDis.BNo4 = dlarray(zeros(fcd*256,1,'single'));
paramsDis.BNs4 = dlarray(ones(fcd*256,1,'single'));
paramsDis.FCW1 = dlarray(initializeGaussian([1,fcd*256*px^3/4096]));
paramsDis.FCb1 = dlarray(zeros(1,1,'single'));
stDis.BN1 = []; stDis.BN2 = []; stDis.BN3 = []; stDis.BN4 = [];

% average Gradient and average Gradient squared holders
avgG.Dis = []; avgGS.Dis = []; avgG.Gen = []; avgGS.Gen = [];
%% Train

if 0 %pretrained loading
disp('Loading Pre-trained initial state, please wait...')
%load('lastepochPTB7') 
epoch = 150; 
settings.maxepochs=200;
end

numIterations = floor(size(trainX,5)/settings.batch_size);
disp("Number of iterations per epoch: "+numIterations)
disp("Calculating total time...")

hf = waitbar(0,'0','Name','Epoch progress... ');
out = false; global_iter = 0; epoch = 0; 
while ~out
    tic; 
    trainXshuffle = trainX(:,:,:,:,randperm(size(trainX,5)));
    for i=1:numIterations
        if rem(i,20)==0 %if too slow increase value
        save('lastiterationPTB') %safety save in case of crashes 
        end
        waitbar(i/numIterations,hf,sprintf('%5.2f%c',i/numIterations*100,'%'))
        global_iter = global_iter+1;
        noise = gpdl(randn([settings.latent_dim,...
            settings.batch_size]),'CB');
        idx = (i-1)*settings.batch_size+1:i*settings.batch_size;
        XBatch=gpdl(single(trainXshuffle(:,:,:,:,idx)),'SSSCB');

        [GradGen,GradDis,stGen,stDis] = ...
                dlfeval(@modelGradients,XBatch,noise,...
                paramsGen,paramsDis,stGen,stDis,px,epoch,i,k_gt);

        % Update Discriminator network parameters
        [paramsDis,avgG.Dis,avgGS.Dis] = ...
            adamupdate(paramsDis, GradDis, ...
            avgG.Dis, avgGS.Dis, global_iter, ...
            settings.lrD, settings.beta1, settings.beta2);

        % Update Generator network parameters
        [paramsGen,avgG.Gen,avgGS.Gen] = ...
            adamupdate(paramsGen, GradGen, ...
            avgG.Gen, avgGS.Gen, global_iter, ...
            settings.lrG, settings.beta1, settings.beta2);

        if i==1 || rem(i,44)==0
            progressplot(paramsGen,stGen,settings,epoch,i,px);
            if 0 % A.A. print .GIF (change to zero to avoid printing)
                if i==1 || (epoch>=0 && i==1) 
                h = gcf;
                % Capture the plot as an image 
                frame = getframe(h); 
                im = frame2im(frame); 
                [imind,cm] = rgb2ind(im,256); 
                % Write to the GIF File 
                    if epoch == 1
                      imwrite(imind,cm,'GANmnist.gif','gif', 'Loopcount',inf); 
                    else 
                      imwrite(imind,cm,'GANmnist.gif','gif','WriteMode','append'); 
                    end 
                end
            end
        end
    end

    elapsedTime = toc;
    if epoch==0
        ett=settings.maxepochs*toc/3600;
        ettm=ett-24;
        disp("Estimated total time: "+ett+" h or one day"+ettm+" h")
        disp("Training...")
    end
    disp("Epoch "+epoch+". Time taken for epoch = "+elapsedTime + "s")    
    epoch = epoch+1;
% For troubleshooting and restarts
%     if epoch==1
%     save('lastepochPTB1')
%     elseif epoch==5
%     save('lastepochPTB5')    
%     elseif epoch==10
%     save('lastepochPTB10')
%     elseif epoch==50
%     save('lastepochPTB50')
%     elseif epoch==100
%     save('lastepochPTB100')
%     elseif epoch==150
%     save('lastepochPTB150')
%     end

    if epoch == settings.maxepochs
        out = true;
    end    
end
%% Helper Functions
%% preprocess
function x = preprocess(x,px)
x = (2^16-1)-double(x); %invert images (sacrifice template technique)
x = double(x)/(2^16-1); %/255; check 16 bit or 8 bit
x = (x-.5)/.5;
x = reshape(x,px,px,px,1,[]);
end
%% extract data
function x = gatext(x)
x = gather(extractdata(x));
end
%% gpu dl array wrapper
function dlx = gpdl(x,labels)
dlx = gpuArray(dlarray(x,labels));
end
%% Weight initialization
function parameter = initializeGaussian(parameterSize,sigma)
if nargin < 2
    sigma = 0.05;
end
parameter = randn(parameterSize, 'single') .* sigma;
end
%% Generator
function [dly,st] = Generator(dlx,params,st,px)
fcg=1/1;
% fully connected
dly = fullyconnect(dlx,params.FCW1,params.FCb1);
%dly = leakyrelu(dly,0.2);
dly = relu(dly);
% transposed convolution
dly = gpdl(reshape(dly,px/8,px/8,px/8,fcg*128,[]),'SSSCB');
dly = dltranspconv(dly,params.TCW1,params.TCb1,...
    'Stride',2,'Cropping','same');
if isempty(st.BN1)
    [dly,st.BN1.mu,st.BN1.sig] = batchnorm(dly,...
        params.BNo1,params.BNs1,'MeanDecay',0.8);
else
    [dly,st.BN1.mu,st.BN1.sig] = batchnorm(dly,params.BNo1,...
        params.BNs1,st.BN1.mu,st.BN1.sig,...
        'MeanDecay',.8);
end
%dly = leakyrelu(dly,0.2);
dly = relu(dly);
dly = dltranspconv(dly,params.TCW2,params.TCb2,...
    'Stride',2,'Cropping','same');
if isempty(st.BN2)
    [dly,st.BN2.mu,st.BN2.sig] = batchnorm(dly,...
        params.BNo2,params.BNs2,'MeanDecay',0.8);
else
    [dly,st.BN2.mu,st.BN2.sig] = batchnorm(dly,params.BNo2,...
        params.BNs2,st.BN2.mu,st.BN2.sig,...
        'MeanDecay',.8);
end
% %dly = leakyrelu(dly,0.2);
dly = relu(dly);
dly = dltranspconv(dly,params.TCW3,params.TCb3,...
    'Stride',2,'Cropping','same');
if isempty(st.BN3)
    [dly,st.BN3.mu,st.BN3.sig] = batchnorm(dly,...
        params.BNo3,params.BNs3,'MeanDecay',0.8);
else
    [dly,st.BN3.mu,st.BN3.sig] = batchnorm(dly,params.BNo3,...
        params.BNs3,st.BN3.mu,st.BN3.sig,...
        'MeanDecay',.8);
end
%dly = leakyrelu(dly,0.2);
dly = relu(dly);
dly = dlconv(dly,params.CNW1,params.CNb1,...
            'Padding','same');
% tanh
dly = tanh(dly);
end
%% Discriminator
function [dly,st] = Discriminator(dlx,params,st,px)
fcd=1/1;
% convolution
%1
dly = dlconv(dlx,params.CNW1,params.CNb1,...
            'Stride',2,'Padding','same');
dly = leakyrelu(dly,0.15);
dly = dropout(dly);
%2
dly = dlconv(dly,params.CNW2,params.CNb2,...
            'Stride',2,'Padding','same');
if isempty(st.BN1)
    [dly,st.BN1.mu,st.BN1.sig] = batchnorm(dly,...
        params.BNo1,params.BNs1,'MeanDecay',0.8);
else
    [dly,st.BN1.mu,st.BN1.sig] = batchnorm(dly,params.BNo1,...
        params.BNs1,st.BN1.mu,st.BN1.sig,...
        'MeanDecay',0.8);
end
dly = leakyrelu(dly,0.15);
dly = dropout(dly);
%3
dly = dlconv(dly,params.CNW3,params.CNb3,...
            'Stride',2,'Padding','same');
if isempty(st.BN2)
    [dly,st.BN2.mu,st.BN2.sig] = batchnorm(dly,...
        params.BNo2,params.BNs2,'MeanDecay',0.8);
else
    [dly,st.BN2.mu,st.BN2.sig] = batchnorm(dly,params.BNo2,...
        params.BNs2,st.BN2.mu,st.BN2.sig,...
        'MeanDecay',0.8);
end
dly = leakyrelu(dly,0.15);
dly = dropout(dly);
%4
dly = dlconv(dly,params.CNW4,params.CNb4,...
            'Stride',2,'Padding','same');
if isempty(st.BN3)
    [dly,st.BN3.mu,st.BN3.sig] = batchnorm(dly,...
        params.BNo3,params.BNs3,'MeanDecay',0.8);
else
    [dly,st.BN3.mu,st.BN3.sig] = batchnorm(dly,params.BNo3,...
        params.BNs3,st.BN3.mu,st.BN3.sig,...
        'MeanDecay',0.8);
end
dly = leakyrelu(dly,0.15);
dly = dropout(dly);
%5
dly = dlconv(dly,params.CNW5,params.CNb5,...
            'Stride',1,'Padding','same');
if isempty(st.BN4)
    [dly,st.BN4.mu,st.BN4.sig] = batchnorm(dly,...
        params.BNo4,params.BNs4,'MeanDecay',0.8);
else
    [dly,st.BN4.mu,st.BN4.sig] = batchnorm(dly,params.BNo4,...
        params.BNs4,st.BN4.mu,st.BN4.sig,...
        'MeanDecay',0.8);
end
dly = leakyrelu(dly,0.15);
dly = dropout(dly);
% Fully connected
dly = gpdl(reshape(dly,fcd*256*px^3/4096,[]),'CB');
dly = fullyconnect(dly,params.FCW1,params.FCb1);
% sigmoid
dly = sigmoid(dly);
end
%% modelGradients
function [GradGen,GradDis,stGen,stDis]=modelGradients(x,z,paramsGen,...
    paramsDis,stGen,stDis,px,epoch,it,k_gt)
[fake_images,stGen] = Generator(z,paramsGen,stGen,px);
d_output_real = Discriminator(x,paramsDis,stDis,px);
[d_output_fake,stDis] = Discriminator(fake_images,paramsDis,stDis,px);

d_pfake=1;
% Loss due to true or not
d_loss = -mean(.9*log(d_output_real+eps)+log(1-d_output_fake+eps));

infs=150; %149; %start FEM information after this epoch
if epoch<infs 
g_loss = -mean(log(d_output_fake+eps));
else
    if it==1 | rem(it,10)==0 %it == 1 %| it == 55 | it == 111 | it == 166 |  it == 27 | it == 82 | it == 138 | it == 192 %compute FEM only in this iterations
        infe=epoch-infs;
        msde=2;   %largest side of element/voxel size in the FEM mesh
        bs=size(x,5);
        kijkl=zeros(6,6,bs);
        for i=1:bs %loop over the batch
            this_is_FEMloop=i
            xrs=(fake_images*0.5+0.5)*(2^16-1);
            kijkl(:,:,i)=FEMmINTNV(xrs(:,:,:,:,i),msde,infe,it,i,30000);
        end
        %computation of rank-sum p-values
        rsk=zeros(6,6);
        k_gtvec11=reshape(k_gt(1,1,:),size(k_gt,3),[]); % diagonal 1
        k_gtvec22=reshape(k_gt(2,2,:),size(k_gt,3),[]); % diagonal 2
        k_gtvec33=reshape(k_gt(3,3,:),size(k_gt,3),[]); % diagonal 3
        k_gtvec44=reshape(k_gt(4,4,:),size(k_gt,3),[]); % diagonal 4
        k_gtvec55=reshape(k_gt(5,5,:),size(k_gt,3),[]); % diagonal 5
        k_gtvec66=reshape(k_gt(6,6,:),size(k_gt,3),[]); % diagonal 6
        k_gtvec12=reshape(k_gt(1,2,:),size(k_gt,3),[]); %1-2
        k_gtvec23=reshape(k_gt(2,3,:),size(k_gt,3),[]); %2-3
        k_gtvec13=reshape(k_gt(1,3,:),size(k_gt,3),[]); %1-3
        kijkl_vec11=reshape(kijkl(1,1,:),bs,[]); % diagonal 1
        kijkl_vec22=reshape(kijkl(2,2,:),bs,[]); % diagonal 2
        kijkl_vec33=reshape(kijkl(3,3,:),bs,[]); % diagonal 3
        kijkl_vec44=reshape(kijkl(4,4,:),bs,[]); % diagonal 4
        kijkl_vec55=reshape(kijkl(5,5,:),bs,[]); % diagonal 5
        kijkl_vec66=reshape(kijkl(6,6,:),bs,[]); % diagonal 6
        kijkl_vec12=reshape(kijkl(1,2,:),bs,[]); %1-2
        kijkl_vec23=reshape(kijkl(2,3,:),bs,[]); %2-3
        kijkl_vec13=reshape(kijkl(1,3,:),bs,[]); %1-3
        rsk(1,1)=ranksum(k_gtvec11, kijkl_vec11); % diagonal 1
        rsk(2,2)=ranksum(k_gtvec22, kijkl_vec22); % diagonal 2
        rsk(3,3)=ranksum(k_gtvec33, kijkl_vec33); % diagonal 3
        rsk(4,4)=ranksum(k_gtvec44, kijkl_vec44); % diagonal 4
        rsk(5,5)=ranksum(k_gtvec55, kijkl_vec55); % diagonal 5
        rsk(6,6)=ranksum(k_gtvec66, kijkl_vec66); % diagonal 6
        rsk(1,2)=ranksum(k_gtvec12, kijkl_vec12); %1-2
        rsk(2,3)=ranksum(k_gtvec23, kijkl_vec23); %2-3
        rsk(1,3)=ranksum(k_gtvec13, kijkl_vec13); %1-3
        %d_pfake=(sum(sum(rsk.^2))).^0.5/9 %alternative averaging
        d_pfake=(sum(sum(rsk.^1)).^1)/9 %average
    end
    g_loss = mean(((log(d_output_fake+eps)).^2+ ...
                   (log(d_pfake+eps)).^2).^(1/2)); %K6
end

% For each network, calculate the gradients with respect to the loss.
GradGen = dlgradient(g_loss,paramsGen,'RetainData',true);
GradDis = dlgradient(d_loss,paramsDis);
end

%% progressplot
function progressplot(paramsGen,stGen,settings,epoch,i,px)
r = 1; c = 1;
noise = gpdl(randn([settings.latent_dim,r*c]),'CB');
gen_imgs = Generator(noise,paramsGen,stGen,px);
gen_imgs = reshape(gen_imgs,settings.image_size(1),settings.image_size(2), ...
    settings.image_size(3),[]); %3D

 fig = gcf;
 if ~isempty(fig.Children)
     delete(fig.Children)
 end
I3d = gatext(gen_imgs); %3D
gen_imgs2d = gen_imgs(settings.image_size(1)/2,:,:,:);
gen_imgs2d = squeeze(gen_imgs2d);
I2d = imtile(gatext(gen_imgs2d)); %2D

I2d = rescale(I2d);
I3d = rescale(I3d);
%figure(2)
%subplot(1,2,2)
%Icut=zeros(32,32);
%Icut=I(:,settings.image_size(2)/2,:);
imagesc(I2d)
% slice(I,settings.side,settings.side/2,1) %3D

title(['Generated image FEM-GAN, slice z=32, Epoch = ' num2str(epoch)])
shading flat
colormap gray
axis square
colorbar

drawnow;

if 0 %save as png image
str="img2d/px" + px + "epoch" + epoch + "i" + i + ".png";
imwrite(I2d, str)
end

% persistent saveim3d
% I3d=double(I3d)
% I3d=extractdata(I3d)
% I3d=gather(I3d);
% saveim3d(:,:,:,epoch+1,i)=(I3d);

%figure(2)
% isosurface (I3d(:,:,:,1),0.10)
% isosurface (I3d(:,:,:,1),0.75)
% isosurface (I3d(:,:,:,1),1.0)
% colorbar

if epoch == settings.maxepochs-0
%     save(images3d,'saveim3d')
    fig3=figure(6)
    orthosliceViewer(I3d(:,:,:,1),'Slicenumbers', ...
        [settings.side,settings.side/2,1])
    axis square
    colorbar
    set(fig3, 'Position', [900 100 1000 950])
    
    fig3=figure(7)
    slice(I3d(:,:,:,1),settings.side,settings.side/2,1) %3D
    axis square
    colorbar
    set(fig3, 'Position', [900 100 1000 950])

    save('last_genimg','gen_imgs','I3d')
    fig2=figure(8)
    sliceViewer(I3d(:,:,:,1));
    axis square
    colorbar
    set(fig2, 'Position', [0 5 800 1000])
    toc
end

end
%% dropout
function dly = dropout(dlx,p)
if nargin < 2
    p = .3;
end
n = p*10;
mask = randi([1,10],size(dlx));
mask(mask<=n)=0;
mask(mask>n)=1;
dly = dlx.*mask;

end